@@ -6,11 +6,12 @@ use crate::{
     protocol::{
         error::Error as ProtocolError,
         messages::{
-            FetchRequest, FetchRequestPartition, FetchRequestTopic, FetchResponse,
-            FetchResponsePartition, IsolationLevel, ListOffsetsRequest,
-            ListOffsetsRequestPartition, ListOffsetsRequestTopic, ListOffsetsResponse,
-            ListOffsetsResponsePartition, ProduceRequest, ProduceRequestPartitionData,
-            ProduceRequestTopicData, ProduceResponse, NORMAL_CONSUMER,
+            DeleteRecordsRequest, DeleteRecordsResponse, DeleteRequestPartition,
+            DeleteRequestTopic, DeleteResponsePartition, FetchRequest, FetchRequestPartition,
+            FetchRequestTopic, FetchResponse, FetchResponsePartition, IsolationLevel,
+            ListOffsetsRequest, ListOffsetsRequestPartition, ListOffsetsRequestTopic,
+            ListOffsetsResponse, ListOffsetsResponsePartition, ProduceRequest,
+            ProduceRequestPartitionData, ProduceRequestTopicData, ProduceResponse, NORMAL_CONSUMER,
         },
         primitives::*,
         record::{Record as ProtocolRecord, *},
@@ -23,7 +24,44 @@ use std::ops::{ControlFlow, Deref, Range};
 use std::sync::Arc;
 use time::OffsetDateTime;
 use tokio::sync::Mutex;
-use tracing::{error, info};
+use tracing::{error, info, warn};
+
+#[derive(Debug, Clone, Copy, PartialEq, Eq)]
+pub enum Compression {
+    NoCompression,
+    #[cfg(feature = "compression-gzip")]
+    Gzip,
+    #[cfg(feature = "compression-lz4")]
+    Lz4,
+    #[cfg(feature = "compression-snappy")]
+    Snappy,
+    #[cfg(feature = "compression-zstd")]
+    Zstd,
+}
+
+impl Default for Compression {
+    fn default() -> Self {
+        Self::NoCompression
+    }
+}
+
+/// Which type of offset should be requested by [`PartitionClient::get_offset`].
+///
+/// # Timestamp-based Queries
+/// In theory the Kafka API would also support querying an offset based on a timestamp, but the behavior seems to be
+/// semi-defined, unintuitive (even within Apache Kafka) and inconsistent between Apache Kafka and Redpanda. So we
+/// decided to NOT expose this option.
+#[derive(Debug, Clone, Copy, PartialEq, Eq)]
+pub enum OffsetAt {
+    /// Earliest existing record.
+    ///
+    /// This is NOT the earliest produced record but the earliest record that is still kept, i.e. the offset might
+    /// change if records are pruned by Kafka (retention policy) or if the are deleted.
+    Earliest,
+
+    /// The latest existing record.
+    Latest,
+}
 
 /// Many operations must be performed on the leader for a partition
 ///
@@ -64,14 +102,18 @@ impl PartitionClient {
     }
 
     /// Produce a batch of records to the partition
-    pub async fn produce(&self, records: Vec<Record>) -> Result<Vec<i64>> {
+    pub async fn produce(
+        &self,
+        records: Vec<Record>,
+        compression: Compression,
+    ) -> Result<Vec<i64>> {
         // skip request entirely if `records` is empty
         if records.is_empty() {
             return Ok(vec![]);
         }
 
         let n = records.len() as i64;
-        let request = &build_produce_request(self.partition, &self.topic, records);
+        let request = &build_produce_request(self.partition, &self.topic, records, compression);
 
         maybe_retry(&self.backoff_config, self, "produce", || async move {
             let broker = self.get().await?;
@@ -84,6 +126,12 @@ impl PartitionClient {
     /// Fetch `bytes` bytes of record data starting at sequence number `offset`
     ///
     /// Returns the records, and the current high watermark.
+    ///
+    ///
+    /// # Error Handling
+    /// Fetching records outside the range known the to broker (marked by low and high watermark) will lead to a
+    /// [`ServerError`](Error::ServerError) with [`OffsetOutOfRange`](ProtocolError::OffsetOutOfRange). You may use
+    /// [`get_offset`](Self::get_offset) to determine the offset range.
     pub async fn fetch_records(
         &self,
         offset: i64,
@@ -98,27 +146,63 @@ impl PartitionClient {
         })
         .await?;
 
-        let records = extract_records(partition.records.0)?;
+        // Redpanda never sends OffsetOutOfRange even when it should. "Luckily" it does not support deletions so we can
+        // implement a simple heuristic.
+        if partition.high_watermark.0 < offset {
+            warn!(
+                "This message looks like Redpanda wants to report a OffsetOutOfRange but doesn't."
+            );
+            return Err(Error::ServerError(
+                ProtocolError::OffsetOutOfRange,
+                String::from("Offset out of range"),
+            ));
+        }
+
+        let records = extract_records(partition.records.0, offset)?;
 
         Ok((records, partition.high_watermark.0))
     }
 
-    /// Get high watermark for this partition.
-    pub async fn get_high_watermark(&self) -> Result<i64> {
-        let request = &build_list_offsets_request(self.partition, &self.topic);
+    /// Get offset for this partition.
+    ///
+    /// Note that the value returned by this method should be considered stale data, since:
+    ///
+    /// - **[`OffsetAt::Earliest`]:** Might be change at any time due to the Kafka retention policy or by
+    ///   [deleting records](Self::delete_records).
+    /// - **[`OffsetAt::Latest`]:** Might be change at any time by [producing records](Self::produce).
+    pub async fn get_offset(&self, at: OffsetAt) -> Result<i64> {
+        let request = &build_list_offsets_request(self.partition, &self.topic, at);
 
-        let partition = maybe_retry(
+        let partition = maybe_retry(&self.backoff_config, self, "get_offset", || async move {
+            let response = self.get().await?.request(&request).await?;
+            process_list_offsets_response(self.partition, &self.topic, response)
+        })
+        .await?;
+
+        extract_offset(partition)
+    }
+
+    /// Delete records whose offset is smaller than the given offset.
+    ///
+    /// # Supported Brokers
+    /// Currently this is only supported by Apache Kafka but NOT by Redpanda, see
+    /// <https://github.com/redpanda-data/redpanda/issues/1016>.
+    pub async fn delete_records(&self, offset: i64, timeout_ms: i32) -> Result<()> {
+        let request =
+            &build_delete_records_request(offset, timeout_ms, &self.topic, self.partition);
+
+        maybe_retry(
             &self.backoff_config,
             self,
-            "get_high_watermark",
+            "delete_records",
             || async move {
                 let response = self.get().await?.request(&request).await?;
-                process_list_offsets_response(self.partition, &self.topic, response)
+                process_delete_records_response(&self.topic, self.partition, response)
             },
         )
         .await?;
 
-        extract_high_watermark(partition)
+        Ok(())
     }
 
     /// Retrieve the broker ID of the partition leader
@@ -283,6 +367,7 @@ where
             match error {
                 Error::Request(RequestError::Poisoned(_) | RequestError::IO(_))
                 | Error::Connection(_) => broker_cache.invalidate().await,
+                Error::ServerError(ProtocolError::InvalidReplicationFactor, _) => {}
                 Error::ServerError(ProtocolError::LeaderNotAvailable, _) => {}
                 Error::ServerError(ProtocolError::OffsetNotAvailable, _) => {}
                 Error::ServerError(ProtocolError::NotLeaderOrFollower, _) => {
@@ -304,7 +389,12 @@ where
         .map_err(Error::RetryFailed)?
 }
 
-fn build_produce_request(partition: i32, topic: &str, records: Vec<Record>) -> ProduceRequest {
+fn build_produce_request(
+    partition: i32,
+    topic: &str,
+    records: Vec<Record>,
+    compression: Compression,
+) -> ProduceRequest {
     let n = records.len() as i32;
 
     // TODO: Retry on failure
@@ -341,7 +431,17 @@ fn build_produce_request(partition: i32, topic: &str, records: Vec<Record>) -> P
             last_offset_delta: n - 1,
             is_transactional: false,
             base_sequence: -1,
-            compression: RecordBatchCompression::NoCompression,
+            compression: match compression {
+                Compression::NoCompression => RecordBatchCompression::NoCompression,
+                #[cfg(feature = "compression-gzip")]
+                Compression::Gzip => RecordBatchCompression::Gzip,
+                #[cfg(feature = "compression-lz4")]
+                Compression::Lz4 => RecordBatchCompression::Lz4,
+                #[cfg(feature = "compression-snappy")]
+                Compression::Snappy => RecordBatchCompression::Snappy,
+                #[cfg(feature = "compression-zstd")]
+                Compression::Zstd => RecordBatchCompression::Zstd,
+            },
             timestamp_type: RecordBatchTimestampType::CreateTime,
             producer_id: -1,
             producer_epoch: -1,
@@ -460,7 +560,10 @@ fn process_fetch_response(
     Ok(response_partition)
 }
 
-fn extract_records(partition_records: Vec<RecordBatch>) -> Result<Vec<RecordAndOffset>> {
+fn extract_records(
+    partition_records: Vec<RecordBatch>,
+    request_offset: i64,
+) -> Result<Vec<RecordAndOffset>> {
     let mut records = vec![];
 
     for batch in partition_records {
@@ -472,6 +575,13 @@ fn extract_records(partition_records: Vec<RecordBatch>) -> Result<Vec<RecordAndO
                 records.reserve(protocol_records.len());
 
                 for record in protocol_records {
+                    let offset = batch.base_offset + record.offset_delta as i64;
+                    if offset < request_offset {
+                        // Kafka does not split record batches on the server side, so we need to do this filtering on
+                        // the client side
+                        continue;
+                    }
+
                     let timestamp = OffsetDateTime::from_unix_timestamp_nanos(
                         (batch.first_timestamp + record.timestamp_delta) as i128 * 1_000_000,
                     )
@@ -490,7 +600,7 @@ fn extract_records(partition_records: Vec<RecordBatch>) -> Result<Vec<RecordAndO
                                 .collect(),
                             timestamp,
                         },
-                        offset: batch.base_offset + record.offset_delta as i64,
+                        offset,
                     })
                 }
             }
@@ -500,7 +610,12 @@ fn extract_records(partition_records: Vec<RecordBatch>) -> Result<Vec<RecordAndO
     Ok(records)
 }
 
-fn build_list_offsets_request(partition: i32, topic: &str) -> ListOffsetsRequest {
+fn build_list_offsets_request(partition: i32, topic: &str, at: OffsetAt) -> ListOffsetsRequest {
+    let timestamp = match at {
+        OffsetAt::Earliest => -2,
+        OffsetAt::Latest => -1,
+    };
+
     ListOffsetsRequest {
         replica_id: NORMAL_CONSUMER,
         isolation_level: Some(IsolationLevel::ReadCommitted),
@@ -508,8 +623,7 @@ fn build_list_offsets_request(partition: i32, topic: &str) -> ListOffsetsRequest
             name: String_(topic.to_owned()),
             partitions: vec![ListOffsetsRequestPartition {
                 partition_index: Int32(partition),
-                // latest offset
-                timestamp: Int64(-1),
+                timestamp: Int64(timestamp),
                 max_num_offsets: Some(Int32(1)),
             }],
         }],
@@ -551,7 +665,7 @@ fn process_list_offsets_response(
     }
 }
 
-fn extract_high_watermark(partition: ListOffsetsResponsePartition) -> Result<i64> {
+fn extract_offset(partition: ListOffsetsResponsePartition) -> Result<i64> {
     match (
         partition.old_style_offsets.as_ref(),
         partition.offset.as_ref(),
@@ -574,3 +688,59 @@ fn extract_high_watermark(partition: ListOffsetsResponsePartition) -> Result<i64
         _ => unreachable!(),
     }
 }
+
+fn build_delete_records_request(
+    offset: i64,
+    timeout_ms: i32,
+    topic: &str,
+    partition: i32,
+) -> DeleteRecordsRequest {
+    DeleteRecordsRequest {
+        topics: vec![DeleteRequestTopic {
+            name: String_(topic.to_string()),
+            partitions: vec![DeleteRequestPartition {
+                partition_index: Int32(partition),
+                offset: Int64(offset),
+                tagged_fields: None,
+            }],
+            tagged_fields: None,
+        }],
+        timeout_ms: Int32(timeout_ms),
+        tagged_fields: None,
+    }
+}
+
+fn process_delete_records_response(
+    topic: &str,
+    partition: i32,
+    response: DeleteRecordsResponse,
+) -> Result<DeleteResponsePartition> {
+    let response_topic = response
+        .topics
+        .exactly_one()
+        .map_err(Error::exactly_one_topic)?;
+
+    if response_topic.name.0 != topic {
+        return Err(Error::InvalidResponse(format!(
+            "Expected data for topic '{}' but got data for topic '{}'",
+            topic, response_topic.name.0
+        )));
+    }
+
+    let response_partition = response_topic
+        .partitions
+        .exactly_one()
+        .map_err(Error::exactly_one_partition)?;
+
+    if response_partition.partition_index.0 != partition {
+        return Err(Error::InvalidResponse(format!(
+            "Expected data for partition {} but got data for partition {}",
+            partition, response_partition.partition_index.0
+        )));
+    }
+
+    match response_partition.error {
+        Some(err) => Err(Error::ServerError(err, String::new())),
+        None => Ok(response_partition),
+    }
+}
