@@ -72,8 +72,8 @@
 //!
 //! // produce data
 //! let record = Record {
-//!     key: b"".to_vec(),
-//!     value: b"hello kafka".to_vec(),
+//!     key: None,
+//!     value: Some(b"hello kafka".to_vec()),
 //!     headers: BTreeMap::from([
 //!         ("foo".to_owned(), b"bar".to_vec()),
 //!     ]),
@@ -145,8 +145,8 @@
 //!         let data = std::mem::take(&mut self.data);
 //!         let records = vec![
 //!             Record {
-//!                 key: b"".to_vec(),
-//!                 value: data,
+//!                 key: None,
+//!                 value: Some(data),
 //!                 headers: BTreeMap::from([
 //!                     ("foo".to_owned(), b"bar".to_vec()),
 //!                 ]),
@@ -207,6 +207,8 @@ use crate::client::producer::aggregator::TryPush;
 use crate::client::{error::Error as ClientError, partition::PartitionClient};
 use crate::record::Record;
 
+use super::partition::Compression;
+
 pub mod aggregator;
 mod broadcast;
 
@@ -230,6 +232,8 @@ pub struct BatchProducerBuilder {
     client: Arc<dyn ProducerClient>,
 
     linger: Duration,
+
+    compression: Compression,
 }
 
 impl BatchProducerBuilder {
@@ -243,6 +247,7 @@ impl BatchProducerBuilder {
         Self {
             client,
             linger: Duration::from_millis(5),
+            compression: Compression::default(),
         }
     }
 
@@ -251,12 +256,21 @@ impl BatchProducerBuilder {
         Self { linger, ..self }
     }
 
+    /// Sets compression.
+    pub fn with_compression(self, compression: Compression) -> Self {
+        Self {
+            compression,
+            ..self
+        }
+    }
+
     pub fn build<A>(self, aggregator: A) -> BatchProducer<A>
     where
         A: aggregator::Aggregator,
     {
         BatchProducer {
             linger: self.linger,
+            compression: self.compression,
             client: self.client,
             inner: Mutex::new(ProducerInner {
                 aggregator,
@@ -268,12 +282,20 @@ impl BatchProducerBuilder {
 
 // A trait wrapper to allow mocking
 trait ProducerClient: std::fmt::Debug + Send + Sync {
-    fn produce(&self, records: Vec<Record>) -> BoxFuture<'_, Result<Vec<i64>, ClientError>>;
+    fn produce(
+        &self,
+        records: Vec<Record>,
+        compression: Compression,
+    ) -> BoxFuture<'_, Result<Vec<i64>, ClientError>>;
 }
 
 impl ProducerClient for PartitionClient {
-    fn produce(&self, records: Vec<Record>) -> BoxFuture<'_, Result<Vec<i64>, ClientError>> {
-        Box::pin(self.produce(records))
+    fn produce(
+        &self,
+        records: Vec<Record>,
+        compression: Compression,
+    ) -> BoxFuture<'_, Result<Vec<i64>, ClientError>> {
+        Box::pin(self.produce(records, compression))
     }
 }
 
@@ -293,6 +315,8 @@ where
 {
     linger: Duration,
 
+    compression: Compression,
+
     client: Arc<dyn ProducerClient>,
 
     inner: Mutex<ProducerInner<A>>,
@@ -370,7 +394,7 @@ where
                 TryPush::NoCapacity(data) => {
                     debug!("Insufficient capacity in aggregator - flushing");
 
-                    Self::flush_impl(&mut inner, self.client.as_ref()).await;
+                    Self::flush_impl(&mut inner, self.client.as_ref(), self.compression).await;
                     match inner
                         .aggregator
                         .try_push(data)
@@ -413,21 +437,28 @@ where
         debug!("Linger expired - flushing");
 
         // Flush data
-        Self::flush_impl(&mut inner, self.client.as_ref()).await;
+        Self::flush_impl(&mut inner, self.client.as_ref(), self.compression).await;
 
-        extract(&result_slot.now_or_never().expect("just flushed"), tag)
+        // We need to poll the result slot here to make the result available via `peek`, otherwise the next thread in
+        // this critical section will flush the producer a 2nd time. We are using an ordinary `.await` call here instead
+        // of `now_or_never` because tokio might preempt us (or to be precise: might preempt polling from the result slot).
+        extract(&result_slot.await, tag)
     }
 
     /// Flushed out data from aggregator.
     pub async fn flush(&self) {
         let mut inner = self.inner.lock().await;
         debug!("Manual flush");
-        Self::flush_impl(&mut inner, self.client.as_ref()).await;
+        Self::flush_impl(&mut inner, self.client.as_ref(), self.compression).await;
     }
 
     /// Flushes out the data from the aggregator, publishes the result to the result slot,
     /// and creates a fresh result slot for future writes to use
-    async fn flush_impl(inner: &mut ProducerInner<A>, client: &dyn ProducerClient) {
+    async fn flush_impl(
+        inner: &mut ProducerInner<A>,
+        client: &dyn ProducerClient,
+        compression: Compression,
+    ) {
         trace!("Flushing batch producer");
 
         let (output, status_deagg) = match inner.aggregator.flush() {
@@ -449,7 +480,7 @@ where
             // "just flushed" because no data is available right after flushing.
             Ok(vec![])
         } else {
-            client.produce(output).await
+            client.produce(output, compression).await
         };
 
         // Reset result slot
@@ -489,7 +520,11 @@ mod tests {
     }
 
     impl ProducerClient for MockClient {
-        fn produce(&self, records: Vec<Record>) -> BoxFuture<'_, Result<Vec<i64>, ClientError>> {
+        fn produce(
+            &self,
+            records: Vec<Record>,
+            _compression: Compression,
+        ) -> BoxFuture<'_, Result<Vec<i64>, ClientError>> {
             Box::pin(async move {
                 tokio::time::sleep(self.delay).await;
 
@@ -510,8 +545,8 @@ mod tests {
 
     fn record() -> Record {
         Record {
-            key: vec![0; 4],
-            value: vec![0; 6],
+            key: Some(vec![0; 4]),
+            value: Some(vec![0; 6]),
             headers: Default::default(),
             timestamp: OffsetDateTime::from_unix_timestamp(320).unwrap(),
         }
